/*
 *	Author: Cheng Long
 *	Email: clong@cse.ust.hk
 */

/*
 *	IR-tree.h defines the structures for implementing the IR-tree.
 *
 *	1. The "delete" operation of the IR-tree is not mentioned in the Cao et al.'s paper,
 *	since this operation is used in the algorithms.
 *	We also implement this operation here.
 *
 *	2. Possible ways for implementing the "delete" operatoin:
 *	When an obj/sub-tree is removed, maintain the set of keywords 
 *	that contained in this obj/sub-tree but not contained by other entries of the same node.
 */

#ifndef IRTREE_H
#define IRTREE_H

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <float.h>


//#include "bit_operator.h"
//#include "bst.h"
//#include "data_struct.h"
//#include "data_struct3.h"
//#include "data_utility.h"

//#define R_tree_OR_R_STAR_TREE	1

#define tree_M				32
#define tree_m				16
#define SPLIT_OPT			2

#define MAX_LINE_LENG	1027
#define MAX_FILENAME_LENG 256

//#define DATA_FILE		"R_tree_DataFile.txt"		//
#define SAVE_FILE		"R_tree.txt"				//savefile of rtree.





#define max(a, b) ((a) >= (b) ? (a) : (b))
#define min(a, b) ((a) <= (b) ? (a) : (b))


//From irtree.h
//The interval structure of one dimension.
typedef struct 
{
	float min;
	float max;

}	range_t;

//The structure of the object.
typedef struct obj
{
	int			id;
	range_t*	MBR;

	//IR-tree augmentation.
	//k_node_t*	k_head;

	//For the application.
	float		weight;
	float*		axis;
	
}	obj_t;

//The structure of storing the data.
typedef struct data		//input data structure.
{
	int		dim;				//Dimension of data object.
	int		n_p_objs;			//The # of objects in P.
	int		n_o_objs;			//The # of objects in O.

	obj_t*	p_objs;				//The array containing P objects.
	obj_t*	o_objs;				//The array containing O objects.

}	data_t;


//The node structure for storing an object pointer.
typedef struct obj_node
{
	obj_t*				obj_v;
	struct obj_node*	next;

	//for const_k_NN_key only.
	//B_KEY_TYPE			dist;
}	obj_node_t;


//The node structure of the node of the rtree.
typedef struct tree_node         
{
	int			num;            //The number of items in the node.
	int			loc;            //The location in its parent's node.

	int			level;			//from 'leaf(0)' to 'root', increased by 1 level by level.
	tree_node*	parent;			//pointer to the parent node.
	void**		child;			//pointer array(because the inner-node and leaf-node has different chidren) 
	range_t**	MBRs;			//closing MBRangle within 'x' dimension.

	//Extension for IR-tree.
	//BIT_TYPE *keys;
		//bit array is adopted for representing the posting list.
		//adjacent list is another data structure option.
	//k_node_t** key;

	//bst_t*	bst_v;
	
}	tree_node_t;

//The info structure of the rtree.
typedef struct	R_tree
{
	int		dim;
	int		split_opt;
	
	tree_node_t*	root;	//root node.

	int		height;			//height.
	int		obj_n;			//the number of the objects indexed by the tree.

	int		inner_n;		//the number of inner-nodes.
	int		leaf_n;			//the number of leaf-nodes.

	//bool*	is_first;		//used for the r*-tree.
	//int	height_max;		//used for the r*-tree.

	//IF info.
	//int		key_n;
	
}	R_tree_t;

//
struct objPointer	//typedef 
{
	obj_t*		p;
	objPointer*	next;
};

//
struct nodePointer	//typedef 
{
	tree_node_t*	p;
	nodePointer*	next;
};



//The structure for storing the configuration info.
typedef	struct
{
//	int		M;
//	int		m;
	int		dim;
	//int	height_max;
	int		obj_n;
	//int		key_n;

	//int		R_M;				
	//int		R_m;

	//int		split_opt;

	//char	data_file[ MAX_FILENAME_LENG];
	char	loc_file[ MAX_FILENAME_LENG];
	//char	doc_file[ MAX_FILENAME_LENG];

	char	tree_file[ MAX_FILENAME_LENG];

	//IR-tree augmentation.

}	R_tree_config_t;


//The data structure for storing the statistics.
typedef struct tree_stat
{
	float		irtree_build_time;

	//memory.
	float		memory_v;			//meaning.
	float		memory_max;			//
	
	float		tree_memory_v;
	float		tree_memory_max;

}	tree_stat_t;



extern R_tree_t R_tree_v;
extern tree_stat_t tree_stat_v;
	

/*---------------------Calculation Related APIs----------------------*/

//double abs( double var);

void CreateNode( tree_node_t* &pnode);

void ReleaseNode( tree_node_t* pnode, int tag);

int IsOverlapped( range_t* MBR1, range_t* MBR2);

bool IsSame( range_t* MBR1, range_t* MBR2);

double GetArea( tree_node_t* pnode);	

double GetArea_2( range_t* MBR1, range_t* MBR2);

double GetIncArea( tree_node_t* pnode, range_t* MBR);

void NeedLeastArea( tree_node_t* pnode, range_t* obj, int& loc);

void calc_MBR( tree_node_t* pnode, range_t* &MBR);

range_t* get_MBR_node( tree_node_t* pnode, int dim);

bool UpdateMRB( tree_node_t* pnode, range_t* &MBR);
 
/*----------------------Insertion related APIs----------------------*/

void ChooseLeaf( tree_node_t* root, range_t* MBR, tree_node_t* &leaf_node);

void PickSeeds( tree_node_t* pnode, range_t* MBR, int& g1, int& g2);

void LinearPickSeeds( tree_node_t* pnode, range_t* MBR, int& g1, int& g2);

int PickNext( tree_node_t* pnode, tree_node_t* node_1, tree_node_t* node_2, tree_node_t* &node_c, range_t* MBR, bool* done, int rTag);

int LinearPickNext( tree_node_t* pnode, tree_node_t* node_1, tree_node_t* node_2, tree_node_t* &node_c, bool* done, int rTag);

void AssignEntry( tree_node_t* pnode, tree_node_t* node_c, void* obj, range_t* MBR);

void SplitNode( tree_node_t* &pnode, tree_node_t* &node_2, void* obj, int opt);

//void QuadraticSplit( tree_node_t* &pnode, tree_node_t* &node_2, void* obj);

//void LinearSplit( tree_node_t* &pnode, tree_node_t* &node_2, void* obj);

void AdjustTree( tree_node_t* pnode);

tree_node_t* CreateNewRoot( tree_node_t* node_1, tree_node_t* node_2);

tree_node_t* InsertSub( tree_node_t* pnode, void* obj);

void Insert( obj_t* obj);

/*---------------------Deletion Related APIs------------------------*/

tree_node_t* Search( tree_node_t* root, range_t* MBR, int& loc);

void ChooseInner( tree_node_t* root, range_t* MBR, tree_node_t* &inner_node);

void CondenseSub( tree_node_t* pnode, objPointer* h1, nodePointer* h2 );

void CondenseTree( tree_node_t* pnode);

int Delete( obj_t* obj); 

/*---------------------Other Operation APIs------------------------*/

void ini_tree( );

bool print_and_check_tree( int o_tag, char* tree_file);

void read_node( FILE* i_fp, tree_node_t* node_v);

obj_t* const_obj( tree_node_t* node_v, int loc, int& id_cnt);

void read_tree( char* ir_tree_f);

void build_R_tree( data_t* data_v);

void free_R_tree_sub( );

void free_R_tree( );

//void bp_free_R_tree( );

void print_MBR( range_t* MBR, int dim);

//void test_R_tree( int o_tag);



//Updated on 27 Feb.
//=====================================



#endif