/*
  This code was written by Raymond Chi-Wing Wong.
  Please read the following.

    Raymond Chi-Wing Wong, Yufei Tao, Ada Wai-Chee Fu and Xiaokui Xiao. 
    "On Efficient Spatial Matching", 
    the 33rd International Conference on Very Large Data Bases (VLDB'07), 
    Vienna, Austria on 23-28 Sept, 2007 

*/

#include "stable.h"


// indexO 
// indexP is not equal to pid
R_TYPE stable_distance(int indexO, int indexP, pointList *O, pointList *P, int noOfDim)
{
	R_TYPE returnValue;
	R_TYPE diff;

	point **listO;
	point **listP;

	point *aPointO;
	point *aPointP;

	R_TYPE *aContentO;
	R_TYPE *aContentP;

	int i;

	// assume the indexO and indexP are correct
	listO = O->list;
	listP = P->list;

	aPointO = listO[indexO];
	aPointP = listP[indexP];

	aContentO = aPointO->content;
	aContentP = aPointP->content;

	returnValue = 0;
	for (i = 0; i < noOfDim; i++)
	{
		diff = aContentO[i] - aContentP[i];
		returnValue += diff*diff;
	}

	return returnValue;
}

void stable_constructTwoPreferenceList(twoData *aTwoData, int ***preferenceListO, int ***preferenceListP, int noOfDim)
{
	pointList *O;
	pointList *P;

	int **indexO;
	int **indexP;

	R_TYPE *tempList;
	int *tempIndex;

	int i, j;

	int noOfPointO, noOfPointP;

	O = aTwoData->O;
	P = aTwoData->P;

	noOfPointO = O->noOfPoint;
	noOfPointP = P->noOfPoint;

	// initailization
	indexO = (int **) malloc(sizeof(int *)*noOfPointO);
	for (i = 0; i < noOfPointO; i++)
	{
		indexO[i] = (int *) malloc(sizeof(int)*noOfPointP);
		memset(indexO[i], 0, sizeof(int)*noOfPointP);
	}
	indexP = (int **) malloc(sizeof(int *)*noOfPointP);
	for (i = 0; i < noOfPointP; i++)
	{
		indexP[i] = (int *) malloc(sizeof(int)*noOfPointO);
		memset(indexP[i], 0, sizeof(int)*noOfPointO);
	}

	// for O
	tempList = (R_TYPE *) malloc(sizeof(R_TYPE)*noOfPointP);
	tempIndex = (int *) malloc(sizeof(int)*noOfPointP);
	for (i = 0; i < noOfPointO; i++)
	{
		memset(tempList, 0, sizeof(R_TYPE)*noOfPointP);
		memset(tempIndex, 0, sizeof(int)*noOfPointP);
		for (j = 0; j < noOfPointP; j++)
		{
			// distance between i and j
			tempList[j] = stable_distance(i, j, O, P, noOfDim);

			// index
			indexO[i][j] = j;
		}

		// sort
#ifdef R_FLOAT
		quicksort(indexO[i], (double*)tempList, 0, noOfPointP-1);
#else
		quicksort_i(indexO[i], tempList, 0, noOfPointP-1);
#endif
		// reverse
		for (j = 0; j < noOfPointP; j++)
		{
			tempIndex[j] = indexO[i][noOfPointP-j-1];
		}
		for (j = 0; j < noOfPointP; j++)
		{
			// different from P (to make the id unique) (do it after qsort)
			indexO[i][j] = tempIndex[j] + noOfPointO;
		}
	}
	free(tempList);
	free(tempIndex);

	

	// for P
	tempList = (R_TYPE *) malloc(sizeof(R_TYPE)*noOfPointO);
	tempIndex = (int *) malloc(sizeof(int)*noOfPointO);
	for (i = 0; i < noOfPointP; i++)
	{
		memset(tempList, 0, sizeof(R_TYPE)*noOfPointO);
		memset(tempIndex, 0, sizeof(int)*noOfPointO);
		for (j = 0; j < noOfPointO; j++)
		{
			// distance between i an j
			tempList[j] = stable_distance(j, i, O, P, noOfDim);

			// index
			indexP[i][j] = j;
		}

		// sort
#ifdef R_FLOAT
		quicksort(indexP[i], ( double*)tempList, 0, noOfPointO-1);
#else
		quicksort_i(indexP[i], tempList, 0, noOfPointO-1);
#endif

		// reverse
		for (j = 0; j < noOfPointO; j++)
		{
			tempIndex[j] = indexP[i][noOfPointO-j-1];
		}
		for (j = 0; j < noOfPointO; j++)
		{
			indexP[i][j] = tempIndex[j];
		}
	}
	free(tempList);
	free(tempIndex);	


	(*preferenceListO) = indexO;
	(*preferenceListP) = indexP;
}


int **stable_constructOneRankList(int **preferenceListP, int noOfPointP, int noOfPointO)
{
	int **returnValue;

	int i, j;

	int aRank;
	int indexO;

	returnValue = (int **) malloc(sizeof(int *)*noOfPointP);
	for (i = 0; i < noOfPointP; i++)
	{
		returnValue[i] = (int *) malloc(sizeof(int)*noOfPointO);
		memset(returnValue[i], 0, sizeof(int)*noOfPointO);

		for (j = 0; j < noOfPointO; j++)
		{
			aRank = j;
			indexO = preferenceListP[i][aRank];

			returnValue[i][indexO] = aRank;
		}
	}

	return returnValue;
}

// aAssignment: the assignment of pid
void stable_insert(int oid, int pid, int weight, assignment *aAssignment, twoData *orgTwoData, twoData *engageTwoData)
{
	int spaceP;

	int qP, engageP;
	point *orgPointP;
	point *orgPointO;
	point *engagePointP;
	point *engagePointO;

	triple *aTriple;

	pointList *engagePointListP;
	pointList *engagePointListO;

	//
	orgPointP = twoData_getPPoint(orgTwoData, pid);
	orgPointO = twoData_getOPoint(orgTwoData, oid);
	engagePointP = twoData_getPPoint(engageTwoData, pid);
	engagePointO = twoData_getOPoint(engageTwoData, oid);

	// space(p) <- q(p) - engage(p)
	qP = orgPointP->weight;
	engageP = engagePointP->weight;
	spaceP = qP - engageP;

	engagePointListP = engageTwoData->P;
	engagePointListO = engageTwoData->O;

	// 
	if (spaceP != 0)
	{
		if (spaceP >= weight)
		{
			// match
			engagePointO->weight += weight;
			engagePointP->weight += weight;

			aTriple = triple_newValue(oid, pid, weight);
			assignment_insert(aAssignment, &aTriple);


		}
		else
		{
			// match
			engagePointO->weight += spaceP;
			engagePointP->weight += spaceP;

			aTriple = triple_newValue(oid, pid, spaceP);
			assignment_insert(aAssignment, &aTriple);
		}

		//
		if (engagePointO->weight == orgPointO->weight)
		{
			engagePointListO->noOfRemain--;
		}
		if (engagePointP->weight == orgPointP->weight)
		{
			engagePointListP->noOfRemain--;
		}
	}
}

void stable_remove(assignment *aAssignment, int index)
{
	int oid, pid, weight;
	triple *aTriple;

	// assume index is correct
	aTriple = aAssignment->list[index];

	oid = aTriple->oid;
	pid = aTriple->pid;
	weight = aTriple->weight;

	assignment_removeIndex(aAssignment, index);

	free(aTriple);
}

void stable_setEngageZero(pointList *aPointList)
{
	int noOfPoint;
	point **list;
	point *aPoint;
	int i;

	noOfPoint = aPointList->noOfPoint;
	list = aPointList->list;

	for (i = 0; i < noOfPoint; i++)
	{
		aPoint = list[i];

		aPoint->weight = 0;
	}
}

int stable_isContainMatch(pointList *engagePointList, pointList *orgPointList)
{
	int returnValue;

	int noOfPoint;

	point **engageList;
	point **orgList;

	point *engagePoint;
	point *orgPoint;

	int i;
	
	noOfPoint = engagePointList->noOfPoint;

	engageList = engagePointList->list;
	orgList = orgPointList->list;

	returnValue = 0;
	for (i = 0; i < noOfPoint; i++)
	{
		engagePoint = engageList[i];
		orgPoint = orgList[i];

		if (engagePoint->weight != orgPoint->weight)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

// NULL: cannot find
assignment *stable_findAssignment(assignmentSet *aAssignmentSetP, int noOfPointO, int noOfPointP, int pid)
{
	assignment *returnValue;

	int noOfAssign;
	assignment **list;

	int index;

	noOfAssign = aAssignmentSetP->noOfAssign;
	list = aAssignmentSetP->list;

	index = pid - noOfPointO;

	if ((0 <= index) && (index < noOfPointP))
	{
		returnValue = list[index];
	}
	else
	{
		returnValue = NULL;
	}

	return returnValue;
}

assignmentSet *stable_oProposeToP(twoData *orgTwoData, twoData *engageTwoData, int *currOPointer, int **preferenceO)
{
	int noOfPointO, noOfPointP;
	int i;

	point **orgList;
	int oid, pid;

	point *engagePointO;
	point *orgPointO;

	pointList *orgO;
	pointList *orgP;

	assignment *aAssignment;
	triple *aTriple;

	int aIndex;

	assignmentSet *aProposalListP;

	orgO = orgTwoData->O;
	orgP = orgTwoData->P;

	noOfPointO = orgO->noOfPoint;
	noOfPointP = orgP->noOfPoint;

	orgList = orgO->list;

	aProposalListP = assignmentSet_new();
	aProposalListP->noOfAssign = noOfPointP;
	aProposalListP->list = (assignment **) malloc(sizeof(assignment *)*noOfPointP);
	for (i = 0; i < noOfPointP; i++)
	{
		aProposalListP->list[i] = assignment_new();
	}

	for (i = 0; i < noOfPointO; i++)
	{
		oid = i;

		engagePointO = twoData_getOPoint(engageTwoData, oid);
		orgPointO = twoData_getOPoint(orgTwoData, oid);

		if (orgPointO->weight > engagePointO->weight)
		{
			// find pid
			aIndex = currOPointer[oid];
			pid = preferenceO[oid][aIndex];

			aAssignment = stable_findAssignment(aProposalListP, noOfPointO, noOfPointP, pid);

			aTriple = triple_newValue(oid, pid, orgPointO->weight - engagePointO->weight);
			assignment_insert(aAssignment, &aTriple);

			currOPointer[oid]++;
		}
	}

	//assignmentSet_free(&aProposalListP);

	

	return aProposalListP;
}

// -1: if aProposal contains nothing
int stable_higestOWithRespectToP(int pid, assignment *aProposal, int *rankP)
{
	int returnValue;

	int noOfMatch;
	triple **list;

	int oid;

	triple *aTriple;

	int aRank, minRank;

	int i;

	noOfMatch = aProposal->noOfMatch;
	list = aProposal->list;

	if (noOfMatch > 0)
	{
		i = 0;
		aTriple = list[i];
		oid = aTriple->oid;
		aRank = rankP[oid];

		minRank = aRank;
		returnValue = oid;

		for (i = 1; i < noOfMatch; i++)
		{
			aTriple = list[i];
			oid = aTriple->oid;
			aRank = rankP[oid];

			if (aRank < minRank)
			{
				minRank = aRank;
				returnValue = oid;
			}
		}
	}
	else
	{
		returnValue = -1;
	}


	return returnValue;
}

//indexSet = stable_findMatchOfPWithLowerPrefThanO(pid, oid, currentMatch);
mySet *stable_findMatchOfPWithLowerPrefThanO(int pid, int oid, assignment *currentMatch, int *rankP)
{
	mySet *returnValue;

	int compareRankO;

	int noOfMatch;
	triple **list;
	triple *aTriple;

	int aRank;

	int oid2;

	int i;

	returnValue = mySet_new();

	noOfMatch = currentMatch->noOfMatch;
	list = currentMatch->list;

	compareRankO = rankP[oid];

	for (i = 0; i < noOfMatch; i++)
	{
		aTriple = list[i];
		oid2 = aTriple->oid;
		aRank = rankP[oid2];

		if (aRank > compareRankO)
		{
			mySet_insert(returnValue, i);
		}
	}

	return returnValue;
}

assignment *stable_removeMatch(assignment *aCurrentMatch, mySet *indexSet, twoData *engageTwoData)
{
	assignment *returnValue;

	int noOfElt;
	int *elt;

	triple *aTriple;

	int i;

	int aIndex;

	int oid, pid, weight;

	triple **matchList;

	point *engageO;
	point *engageP;

	

	noOfElt = indexSet->noOfElt;
	elt = indexSet->elt;

	returnValue = assignment_new();

	for (i = noOfElt-1; i >= 0; i--)
	{
		aIndex = elt[i];

		// add
		matchList = aCurrentMatch->list;
		aTriple = matchList[aIndex];
		assignment_insert(returnValue, &aTriple);

		// remove
		oid = aTriple->oid;
		pid = aTriple->pid;
		weight = aTriple->weight;
		engageO = twoData_getOPoint(engageTwoData, oid);
		engageP = twoData_getPPoint(engageTwoData, pid);

		engageO->weight -= weight;
		engageP->weight -= weight;
		assignment_removeIndex(aCurrentMatch, aIndex);

		//
		//orgO = twoData_getOPoint(orgTwoData, oid);
		//orgP = twoData_getPPoint(orgTwoData, pid);
	}

	return returnValue;
}


// NULL : when aAssignment contains 0 matches.
int *stable_sortO(assignment *aAssignment, int *rankP)
{
	int *returnValue;
	int *value;
	int *index;

	int noOfMatch;
	triple **list;

	int i;

	int oid;
	int aRank;
	triple *aTriple;

	noOfMatch = aAssignment->noOfMatch;
	list = aAssignment->list;

	if (noOfMatch == 0)
	{
		returnValue = NULL;
	}
	else
	{
		index = (int *) malloc(sizeof(int)*noOfMatch);
		returnValue = (int *) malloc(sizeof(int)*noOfMatch);
		value = (int *) malloc(sizeof(int)*noOfMatch);
		memset(index, 0, sizeof(int)*noOfMatch);
		memset(returnValue, 0, sizeof(int)*noOfMatch);
		memset(value, 0, sizeof(int)*noOfMatch);

		for (i = 0; i < noOfMatch; i++)
		{
			index[i] = i;

			aTriple = list[i];
			oid = aTriple->oid;
			aRank = rankP[oid];

			value[i] = aRank;
		}

		quicksort_i(index, value, 0, noOfMatch - 1);

		// reverse the returnValue
		for (i = 0; i < noOfMatch; i++)
		{
			returnValue[i] = index[noOfMatch-i-1];
		}

		free(value);
		free(index);
	}


	return returnValue;
}

void stable_pReceiveOProposal(twoData *orgTwoData, twoData *engageTwoData, assignmentSet *aProposalListP, assignmentSet *aCurrentMatchSet, int **rankP)
{
	pointList *orgO;
	pointList *orgP;

	int i, j;

	int pid, oid, weight;

	triple *aTriple;

	int noOfPointO, noOfPointP;

	mySet *indexSet;

	int aIndex;

	int *index;

	assignment *aProposal;
	assignment *aCurrentMatch;

	assignment *o2Match;
	assignment *o3Match;

	orgO = orgTwoData->O;
	orgP = orgTwoData->P;

	noOfPointO = orgO->noOfPoint;
	noOfPointP = orgP->noOfPoint;

	for (i = 0; i < noOfPointP; i++)
	{
		pid = noOfPointO + i;
		aProposal = stable_findAssignment(aProposalListP, noOfPointO, noOfPointP, pid);

		if (aProposal->noOfMatch == 0)
		{
			// do nothing
		}
		else
		{
			// find highest priority o
			oid = stable_higestOWithRespectToP(pid, aProposal, rankP[i]);

			// find O' (a set of o''s which are the matches of p whre p prefers o to o')
			aCurrentMatch = stable_findAssignment(aCurrentMatchSet, noOfPointO, noOfPointP, pid);
			indexSet = stable_findMatchOfPWithLowerPrefThanO(pid, oid, aCurrentMatch, rankP[i]);

if (i == 433)
{
	printf("");
}
			// remove and get o'
			o2Match = stable_removeMatch(aCurrentMatch, indexSet, engageTwoData);

			// union o' and L to form o''
			o3Match = assignment_union(o2Match, aProposal);

			// sort O'' according to p's preference list
			index = stable_sortO(o3Match, rankP[i]);

			// for each o'' in O'' 
			//  insert(o'', p, w')
			for (j = 0; j < o3Match->noOfMatch; j++)
			{
				aIndex = index[j];

				aTriple = o3Match->list[aIndex];
				oid = aTriple->oid;
				pid = aTriple->pid;
				weight = aTriple->weight;

				stable_insert(oid, pid, weight, aCurrentMatch, orgTwoData, engageTwoData);
			}
			
			mySet_free(&indexSet);

			assignment_free(&o2Match);
			//free o3Match
			if (o3Match->noOfMatch != 0)
			{
				free(o3Match->list);
			}
			free(o3Match);

			if (index != NULL)
			{
				free(index);
			}
		}
	}
}

assignment *stable_collapseFromCurrentMatchSet_and_destroy(assignmentSet **aAssignmentSet)
{
	assignment *returnValue;

	int noOfAssign, noOfMatch;
	assignment **list;

	triple **matchList;

	int i, j;

	triple *aTriple;

	assignment *aAssignment;

	returnValue = assignment_new();

	noOfAssign = (*aAssignmentSet)->noOfAssign;
	list = (*aAssignmentSet)->list;

	for (i = 0; i < noOfAssign; i++)
	{
		aAssignment = list[i];

		noOfMatch = aAssignment->noOfMatch;
		matchList = aAssignment->list;

		for (j = 0; j < noOfMatch; j++)
		{
			aTriple = matchList[j];
			assignment_insert(returnValue, &aTriple);
		}

		if (noOfMatch != 0)
		{
			free(matchList);
		}
	}

	if (noOfAssign != 0)
	{
		free(list);
	}
	free((*aAssignmentSet));

	return returnValue;
}

assignment *stable_GaleShapley(twoData *orgTwoData, int noOfDim, int **preferenceO, int **rankP)
{
	assignment *returnValue;

	assignmentSet *aProposalListP;
	twoData *engageTwoData;

	int *currOPointer;

	int isContainMatchO, isContainMatchP;

	pointList *engageO;
	pointList *engageP;

	pointList *orgO;
	pointList *orgP;

	assignmentSet *aCurrentMatchSet;

	int noOfPointO, noOfPointP;
	int i;

	// clone
	engageTwoData = twoData_clone(orgTwoData, noOfDim);

	engageO = engageTwoData->O;
	engageP = engageTwoData->P;

	orgO = orgTwoData->O;
	orgP = orgTwoData->P;

	noOfPointO = engageO->noOfPoint;
	noOfPointP = engageP->noOfPoint;

	currOPointer = (int *) malloc(sizeof(int)*noOfPointO);
	memset(currOPointer, 0, sizeof(int)*noOfPointO);

	aCurrentMatchSet = assignmentSet_new();

	aCurrentMatchSet->noOfAssign = noOfPointP;
	aCurrentMatchSet->list = (assignment **) malloc(sizeof(assignment *)*noOfPointP);
	for (i = 0; i < noOfPointP; i++)
	{
		aCurrentMatchSet->list[i] = assignment_new();
	}

	// initialize engage
	stable_setEngageZero(engageO);
	stable_setEngageZero(engageP);



	// 
	isContainMatchO = stable_isContainMatch(engageO, orgO);
	isContainMatchP = stable_isContainMatch(engageP, orgP);


	while ((isContainMatchO == 1) && (isContainMatchP == 1))
	{
		// o propose to p
		aProposalListP = stable_oProposeToP(orgTwoData, engageTwoData, currOPointer, preferenceO);

		// p recevies o's proposal
		stable_pReceiveOProposal(orgTwoData, engageTwoData, aProposalListP, aCurrentMatchSet, rankP);

		// free aProposalListP
		assignmentSet_free(&aProposalListP);

		// next iteration
		isContainMatchO = stable_isContainMatch(engageO, orgO);
		isContainMatchP = stable_isContainMatch(engageP, orgP);
	}

	// collapse
	returnValue = stable_collapseFromCurrentMatchSet_and_destroy(&aCurrentMatchSet);

	// free
	free(currOPointer);

	return returnValue;
}


//assignment *stable_main()
